<?php

declare(strict_types=1);
namespace Drupal\sessionless\WebTokenAlgorithms;

use Jose\Component\Core\JWK;
use Jose\Component\Encryption\Algorithm\ContentEncryption\A256GCM;
use Jose\Component\Encryption\Algorithm\ContentEncryptionAlgorithm;
use Jose\Component\Encryption\Algorithm\KeyEncryption\A256GCMKW;
use Jose\Component\Encryption\Algorithm\KeyEncryptionAlgorithm;
use Jose\Component\Encryption\Compression\CompressionMethod;
use Jose\Component\Encryption\Compression\Deflate;
use Jose\Component\KeyManagement\JWKFactory;
use Jose\Component\Signature\Algorithm\HS512;
use Jose\Component\Signature\Algorithm\MacAlgorithm;
use Jose\Component\Signature\Algorithm\SignatureAlgorithm;
use Jose\Easy\Build;

/**
 * Crypto worker.
 *
 * Default algorithms copied over from Eclipse SmallRye, @link https://quarkus.io/guides/security-jwt-build
 *
 * Links:
 * - The easy way: https://web-token.spomky-labs.com/v/v2.x/the-easy-way/easy
 * - SigAlgos: https://web-token.spomky-labs.com/v/v2.x/the-components/signed-tokens-jws/signature-algorithms
 * - EncAlgos: https://web-token.spomky-labs.com/v/v2.x/the-components/encrypted-tokens-jwe/encryption-algorithms
 * - Benchmarks: https://web-token.spomky-labs.com/v/v2.x/benchmark/result-table
 * - RSA: https://connect2id.com/products/nimbus-jose-jwt/examples/jwt-with-rsa-encryption
 * - Signing overview: https://auth0.com/blog/json-web-token-signing-algorithms-overview/
 */
final class WebTokenAlgorithms implements WebTokenAlgorithmsInterface {

  public function __construct() {}

  protected function getSignatureAlgorithm(): SignatureAlgorithm|MacAlgorithm {
    return new HS512();
  }

  public function getSignatureKey(): JWK {
    return JWKFactory::createOctKey(512);
  }

  public function sign(JWK $jwk, array $headers, array $payload): string {
    $jws = Build::jws()
      ->alg($this->getSignatureAlgorithm())
      ->payload($payload);
    foreach ($headers as $headerKey => $headerValue) {
      $jws->header($headerKey, $headerValue);
    }
    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    $signed = $jws->sign($jwk);
    return $signed;
  }


  public function getEncryptionKey(): JWK {
    return JWKFactory::createOctKey(256);
  }

  protected function getKeyEncryptionAlgorithm(): KeyEncryptionAlgorithm {
    return new A256GCMKW();
  }

  protected function getContentEncryptionAlgorithm(): ContentEncryptionAlgorithm {
    return new A256GCM();
  }

  protected function getCompressionMethod(): CompressionMethod {
    return new Deflate(9);
  }

  public function encrypt(JWK $jwk, array $headers, array $payload): string {
    /** @noinspection PhpParamsInspection */
    $jwe = Build::jwe()
      ->alg($this->getKeyEncryptionAlgorithm())
      ->enc($this->getContentEncryptionAlgorithm())
      ->zip($this->getCompressionMethod())
      ->payload($payload);
    foreach ($headers as $headerKey => $headerValue) {
      $jwe->header($headerKey, $headerValue);
    }
    $encrypted = $jwe->encrypt($jwk);
    return $encrypted;
  }

}
