<?php

declare(strict_types=1);
namespace Drupal\sessionless\WebTokenService;

use Drupal\sessionless\KeyStorage\KeyStorageInterface;
use Drupal\sessionless\WebTokenAlgorithms\WebTokenAlgorithmsInterface;
use Jose\Easy\Load;

/**
 * @internal
 */
trait WebTokenServiceTrait {

  protected KeyStorageInterface $keyStorage;
  protected WebTokenAlgorithmsInterface $tokenAlgorithms;

  protected function sign(mixed $data, array $headers = [], array $claims = []): string {
    $jwk = $this->keyStorage->getSignatureKey();
    $payload = ['d' => $data] + $claims;
    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    $signed = $this->tokenAlgorithms->sign($jwk, $headers, $payload);
    return $signed;
  }

  protected function parseSigned(string $token): mixed {
    /** @var \Jose\Easy\JWT $jwt */
    $jwt = Load::jws($token)
      ->key($this->keyStorage->getSignatureKey())
      ->run();
    return $jwt->claims->get('d');
  }

  protected function encrypt(mixed $data, array $headers = [], array $claims = []): string {
    // https://web-token.spomky-labs.com/v/v2.x/benchmark/result-table
    // https://web-token.spomky-labs.com/v/v2.x/the-components/encrypted-tokens-jwe/encryption-algorithms
    $jwk = $this->keyStorage->getEncryptionKey();
    $payload = ['d' => $data] + $claims;
    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    $encrypted = $this->tokenAlgorithms->encrypt($jwk, $headers, $payload);
    return $encrypted;
  }

  protected function parseEncrypted(string $token): mixed {
    /** @var \Jose\Easy\JWT $jwt */
    $jwt = Load::jwe($token)
      ->key($this->keyStorage->getEncryptionKey())
      ->run();
    return $jwt->claims->get('d');
  }

  protected function packData(mixed $data): string {
    $serialized = serialize($data);
    // Serialized data may still contain byte sequences not safe for JSON.
    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    $encoded = base64_encode($serialized);
    return $encoded;
  }

  protected function unpackData(string $encoded): mixed {
    $serialized = base64_decode($encoded);
    if ($serialized === FALSE) {
      throw new \UnexpectedValueException('Invalid encoded data.');
    }
    // Unserializing incoming data here is safe, as we only use signed data.
    $unserialized = @unserialize($serialized);
    if ($unserialized === FALSE && $serialized !== 'b:0;') {
      throw new \UnexpectedValueException('Invalid serialized data.');
    }
    return $unserialized;
  }


}
