<?php

declare(strict_types=1);
namespace Drupal\Tests\sessionless\Unit;

use Drupal\Core\KeyValueStore\KeyValueMemoryFactory;
use Drupal\Core\State\State;
use Drupal\sessionless\KeyStorage\KeyStorage;
use Drupal\sessionless\KeyStorage\KeyStorageInterface;
use Drupal\sessionless\WebTokenAlgorithms\WebTokenAlgorithms;
use Drupal\sessionless\WebTokenAlgorithms\WebTokenAlgorithmsCacheDecorator;
use Drupal\sessionless\WebTokenAlgorithms\WebTokenAlgorithmsInterface;
use Drupal\sessionless\WebTokenService\WebTokenEncryptAndSignService;
use Drupal\sessionless\WebTokenService\WebTokenOnlySignService;
use Drupal\sessionless\WebTokenService\WebTokenServiceInterface;
use Drupal\Tests\sessionless\Traits\ClassWithPrivateProperty;
use Drupal\Tests\sessionless\Traits\SessionlessServices;
use Drupal\Tests\UnitTestCase;
use Drupal\TestTools\Random;
use Symfony\Component\Stopwatch\Stopwatch;

/**
 * @group sessionless
 */
final class WebTokenServiceUnitTest extends UnitTestCase {

  protected function setUp(): void {
    parent::setUp();
  }

  /**
   * @dataProvider provideWebTokenService
   */
  public function testWebTokenService(WebTokenServiceInterface $tokenService): void {
    // Class with private property contains nul bytes when serialized.
    $data = [
      'randomObject' => Random::object(10),
      'classWithPrivateString' => new ClassWithPrivateProperty(Random::string(500)),
    ];
    $token = $tokenService->dump($data);
    $this->assertIsString($token);
    // $this->assertSame('', $token);
    $this->assertMatchesRegularExpression('~^[A-Za-z0-9-_=]+\.[A-Za-z0-9-_=]+\.?[A-Za-z0-9-_.+/=]*$~', $token);
    $recoveredData = $tokenService->parse($token);
    // The recreated object is equal, but not same.
    $this->assertEquals($data, $recoveredData);
    $this->assertNotSame($data, $recoveredData);
  }

  /**
   * @dataProvider provideWebTokenServiceAndKeyStorage
   */
  public function testKeyChange(WebTokenServiceInterface $tokenService, KeyStorageInterface $keyStorage): void {
    $data = random_bytes(100000);

    $token1a = $tokenService->dump($data);
    $token1b = $tokenService->dump($data);

    $keyVersion1 = $keyStorage->getKeyVersion();
    $keyStorage->dropKeys();
    $keyVersion2 = $keyStorage->getKeyVersion();

    $token2a = $tokenService->dump($data);
    $token2b = $tokenService->dump($data);

    // Key version changed.
    $this->assertNotSame($keyVersion1, $keyVersion2);

    // Same key, same token.
    $this->assertSame($token1a, $token1b);
    $this->assertSame($token2a, $token2b);

    // Different key, different token.
    $this->assertNotSame($token1a, $token2a);
  }

  /**
   * @dataProvider provideWebTokenService
   */
  public function testAlgorithmsCaching(WebTokenServiceInterface $tokenService): void {
    $data = random_bytes(100000);
    $stopwatch = new Stopwatch(true);

    $stopwatch->start('1a');
    $tokenService->dump($data);
    $stopwatch->stop('1a');

    $stopwatch->start('1b');
    $tokenService->dump($data);
    $stopwatch->stop('1b');

    // Every second run was significantly faster (but still needs packing).
    $this->assertLessThan($stopwatch->getEvent('1a')->getDuration() / 2, $stopwatch->getEvent('1b')->getDuration());
  }

  public function provideWebTokenService() {
    yield 'onlySigned' => [$this->createWebTokenService(false)->webTokenService];
    yield 'encryptedSigned' => [$this->createWebTokenService(true)->webTokenService];
  }

  public function provideWebTokenServiceAndKeyStorage() {
    $services = $this->createWebTokenService(FALSE);
    yield 'onlySigned' => [$services->webTokenService, $services->keyStorage];
    $services = $this->createWebTokenService(TRUE);
    yield 'encryptedSigned' => [$services->webTokenService, $services->keyStorage];
  }

  public function createWebTokenService(bool $encrypt): SessionlessServices {
    $webTokenAlgorithms = new WebTokenAlgorithms();
    $cachedWebTokenAlgorithms = new WebTokenAlgorithmsCacheDecorator($webTokenAlgorithms);
    $state = new State(new KeyValueMemoryFactory());
    $keyStorage = new KeyStorage($cachedWebTokenAlgorithms, $state);
    if ($encrypt) {
      $tokenService = new WebTokenEncryptAndSignService($keyStorage, $cachedWebTokenAlgorithms);
    }
    else {
      $tokenService = new WebTokenOnlySignService($keyStorage, $cachedWebTokenAlgorithms);
    }
    return new SessionlessServices($tokenService, $cachedWebTokenAlgorithms, $keyStorage);
  }

}
